import importlib
import inspect
import logging
import signal
import threading
from concurrent import futures
from importlib.util import find_spec

import grpc
from prometheus_client import start_http_server
from py_grpc_prometheus.prometheus_server_interceptor import PromServerInterceptor
from sentry_sdk import init as sentry_init

from anilius.core.module import Module
from anilius.core.service import Service
from anilius.core.settings import settings
from anilius.utils.singleton import Singleton

NUM_SECS_TO_WAIT = 10

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


class App(metaclass=Singleton):
    _modules = {}

    def __init__(self):
        self.service = Service()

        self.enable_modules = settings["ENABLE_MODULES"]

        assert (
            type(self.enable_modules) is tuple or type(self.enable_modules) is list
        ), "ENABLE_MODULES should be iterable"

        self._initialize_modules()

        self.event_shutting_down = threading.Event()
        self.server = grpc.server(
            futures.ThreadPoolExecutor(max_workers=10),
            interceptors=(PromServerInterceptor(),),
        )
        start_http_server(8081)
        self.set_signals()
        sentry_init(debug=settings["DEBUG"], dsn=settings["SENTRY_DSN"])

    def set_signals(self):
        signal.signal(signal.SIGTERM, self.on_done)

    def on_done(self, signum, frame):
        self.logger.info("Got signal {}, {}".format(signum, frame))
        self.event_shutting_down.set()

    def _initialize_module(self, module):
        self._modules[module.name] = module()

    def _initialize_modules(self):
        root_of_module = "modules"

        for module in self.enable_modules:
            module_path = root_of_module + "." + module
            print(module_path)

            assert find_spec(module_path) is not None, "Enabled modules should exist"

            mod = importlib.import_module(module_path)

            for entity_key in dir(mod):
                entity = getattr(mod, entity_key)
                if (
                    inspect.isclass(entity)
                    and issubclass(entity, Module)
                    and entity is not Module
                ):
                    self._initialize_module(entity)

    @property
    def logger(self):
        return logger

    def graceful_stop(self):
        self.logger.info("Stopped RPC server, Waiting for RPCs to complete...")
        self.server.stop(NUM_SECS_TO_WAIT).wait()
        self.logger.info("Done stopping server")
        self.server.wait_for_termination()

    def run(self):
        try:
            self.service.add_handlers_to_server(self.server)
            self.server.add_insecure_port("[::]:8080")
            self.server.start()
            self.logger.info("Started server at " + "8080")
            self.event_shutting_down.wait()
            self.graceful_stop()
        except KeyboardInterrupt:
            self.graceful_stop()



from __future__ import print_function

import sys

from anilius.cli.anilius import cli


def main(as_module=False):
    cli.main(args=sys.argv[1:], prog_name="python -m anilius" if as_module else None)


if __name__ == "__main__":
    main(as_module=True)



from abc import ABC

import grpc
from anilius.core.serializer import Serializer

from anilius.core.serializer_field import SerializerField

from anilius.core.permission import Permission


class Controller(ABC):
    permissions = ()
    request_serializer = None
    reply_serializer = None
    _serialized_data = None
    meta = {}

    def __init__(self, request, context):
        self.request = request
        self.context = context
        self.metadata = context.invocation_metadata()

        self.parse()

    def check_permissions(self):
        for permission in self.permissions:
            has_perm = permission.has_perm(self)
            if not has_perm:
                self.raise_permission()
                break

    def raise_permission(self):
        self.context.set_code(grpc.StatusCode.PERMISSION_DENIED)
        self.context.set_details("You have not permission for this action")

    def parse(self):
        self._serialized_data = self.get_request_serializer(self.request).to_dict()

        for data in self.metadata:
            self.meta[data.key] = data.value

        for permission in self.permissions:
            assert isinstance(
                permission, Permission
            ), "permissions should be type of Permission"

    def get_valid_data(self, key, default=None):
        field = self._serialized_data.get(key, None)

        if not isinstance(field, SerializerField):
            return default

        return field.get_value()

    @property
    def get_request_serializer(self):
        return (
            self.request_serializer
            if self.request_serializer is not None
            else Serializer
        )

    @property
    def client_id(self):
        return self.meta.get("client-id", None)

    @property
    def client_secret(self):
        return self.meta.get("client-secret", None)

    @property
    def sdk_id(self):
        return self.meta.get("sdk-secret", None)

    @property
    def sdk_secret(self):
        return self.meta.get("sdk-secret", None)

    @property
    def authorization(self):
        return self.meta.get("authorization ", None)

    def get_response(self):
        self.check_permissions()



"""
Settings and configuration for Django.

Read values from the module specified by the DJANGO_SETTINGS_MODULE environment
variable, and then from django.conf.global_settings; see the global_settings.py
for a list of all possible variables.
"""

import importlib


class Settings(object):
    """
    A lazy proxy for either global Django settings or a custom settings object.
    The user can manually configure settings prior to using them. Otherwise,
    Django uses the settings module pointed to by DJANGO_SETTINGS_MODULE.
    """

    def __init__(self):
        """
        Load the settings module pointed to by the environment variable. This
        is used the first time settings are needed, if the user hasn't
        configured settings manually.
        """
        settings_module = "settings"

        mod = importlib.import_module(settings_module)

        self._explicit_settings = set()
        for setting in dir(mod):
            if setting.isupper():
                setting_value = getattr(mod, setting)
                setattr(self, setting, setting_value)
                self._explicit_settings.add(setting)

    def __getattr__(self, name):
        """Return the value of a setting and cache it in self.__dict__."""
        val = getattr(self, name)
        self.__dict__[name] = val
        return val

    def __getitem__(self, x):
        return getattr(self, x)

    def __setattr__(self, name, value):
        """
        Set the value of setting. Clear all cached values if _wrapped changes
        (@override_settings does this) or clear single values when set.
        """
        self.__dict__.pop(name, None)
        super().__setattr__(name, value)

    def __delattr__(self, name):
        """Delete a setting and clear it from cache if needed."""
        super().__delattr__(name)
        self.__dict__.pop(name, None)


settings = Settings()
